/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Fixed line SMS, to the ETSI protocol
 * 
 * Copyright (C) 2004, Steve Underwood
 *
 * Steve Underwood <steveu@coppice.org>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */
 
#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/translate.h>
#include <asterisk/dsp.h>
#include <asterisk/manager.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <pthread.h>
#include <tiffio.h>

#include <spandsp.h>
#include "/home/steveu/spandsp/dll_sms.h"

static char *tdesc = "Fixed Line SMS, per the ETSI protocol";

static char *app = "SMS";

static char *synopsis = "Exchange SMS messages with a fixed line SMS phone";

static char *descrip = 
"  SMS(message): Exchange an SMS message with a fixed line SMS phone\n"
"Returns -1 when the user hangs up.\n";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

#define BLOCK_SIZE 160

adsi_tx_state_t tx_adsi;

static int adsi_create_message(adsi_tx_state_t *s, uint8_t *msg)
{
    int len;

    len = adsi_add_field(s, msg, -1, CLIP_MDMF_SMS, NULL, 0);
    /* Active */
    len = adsi_add_field(s, msg, len, CLIP_DISPLAY_INFO, "\00ABC", 4);

    return len;
}
/*- End of function --------------------------------------------------------*/

int adsi_tl_next_field(adsi_rx_state_t *s, const uint8_t *msg, int msg_len, int pos, uint8_t *field_type, uint8_t const **field_body, int *field_len)
{
    int len;

    if (pos >= msg_len)
        return -1;
    /* These standards all use "IE" type fields - type, length, body - and similar headers */
    if (pos <= 0)
    {
        /* Return the message type */
        *field_type = msg[0];
        *field_len = 0;
        *field_body = NULL;
        pos = 2;
    }
    else
    {
        *field_type = msg[pos++];
        len = msg[pos++];
        len |= (msg[pos++] << 8);
        *field_len = len;
        *field_body = msg + pos;
        pos += len;
    }
    return pos;
}
/*- End of function --------------------------------------------------------*/

int adsi_tl_add_field(adsi_tx_state_t *s, uint8_t *msg, int len, uint8_t field_type, uint8_t const *field_body, int field_len)
{
    /* These standards all use "IE" type fields - type, length, body - and similar headers */
    if (len <= 0)
    {
        /* Initialise a new message. The field type is actually the message type. */
        msg[0] = field_type;
        msg[1] = 0;
        len = 2;
    }
    else
    {
        /* Add to a message in progress. */
        if (field_type)
        {
            msg[len] = field_type;
            msg[len + 1] = field_len & 0xFF;
            msg[len + 2] = (field_len >> 8) & 0xFF;
            memcpy(msg + len + 3, field_body, field_len);
            len += (field_len + 3);
        }
        else
        {
            /* No field type or length, for restricted single message formats */
            memcpy(msg + len, field_body, field_len);
            len += field_len;
        }
    }
    return len;
}
/*- End of function --------------------------------------------------------*/

int adsi_tl_add_tm(adsi_tx_state_t *s, char *msg)
{
    uint8_t tx_msg[256];
    int tx_len;
    int l;
    
    l = -1;
    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_INFO_MT, NULL, 0);
    tx_len += 2;
    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, tx_len, DLL_PARM_PROVIDER_ID, "AST", 3);
    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, tx_len, DLL_PARM_DESTINATION, "98765432", 8);
    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, tx_len, DLL_PARM_DISPLAY_INFO, msg, strlen(msg));
    tx_msg[2] = tx_len - 4;
    tx_msg[3] = 0;
    adsi_put_message(s, tx_msg, tx_len);
}
/*- End of function --------------------------------------------------------*/

static void put_adsi_msg(void *user_data, const uint8_t *msg, int len)
{
    int i;
    int l;
    uint8_t field_type;
    const uint8_t *field_body;
    int field_len;
    uint8_t body[256];
    adsi_rx_state_t *rx_adsi;
    uint8_t tx_msg[256];
    int tx_len;
    int file;
    
    printf("Good message received (%d bytes)\n", len);

    rx_adsi = (adsi_rx_state_t *) user_data;
    for (i = 0;  i < len;  i++)
    {
        printf("%2x ", msg[i]);
        if ((i & 0xf) == 0xf)
            printf("\n");
    }
    printf("\n");
    l = -1;
    do
    {
        l = adsi_tl_next_field(rx_adsi, msg, len, l, &field_type, &field_body, &field_len);
        if (l > 0)
        {
            if (field_body)
            {
                memcpy(body, field_body, field_len);
                body[field_len] = '\0';
                printf("Type %x, len %d, '%s'\n", field_type, field_len, body);
            }
            else
            {
                printf("Message type %x\n", field_type);
                switch (field_type)
                {
                case DLL_SMS_P2_INFO_MO:
                    file = open("/tmp/pdus", O_WRONLY | O_CREAT, 0666);
                    if (file >= 0)
                    {
                        write(file, msg + 2, msg[1]);
                        close(file);
                    }
                    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_ACK0, NULL, 0);
                    adsi_put_message(&tx_adsi, tx_msg, tx_len);
                    /* Skip the TL length */
                    l += 2;
                    break;
                case DLL_SMS_P2_INFO_MT:
                    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_ACK0, NULL, 0);
                    adsi_put_message(&tx_adsi, tx_msg, tx_len);
                    /* Skip the TL length */
                    l += 2;
                    break;
                case DLL_SMS_P2_INFO_STA:
                    l += 2;
                    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_ACK0, NULL, 0);
                    adsi_put_message(&tx_adsi, tx_msg, tx_len);
                    break;
                case DLL_SMS_P2_NACK:
                    adsi_tl_add_tm(&tx_adsi, "This is the message");
                    l += 2;
                    break;
                case DLL_SMS_P2_ACK0:
                    l += 2;
                    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_ACK0, NULL, 0);
                    adsi_put_message(&tx_adsi, tx_msg, tx_len);
                    break;
                case DLL_SMS_P2_ACK1:
                    l += 2;
                    adsi_tl_add_tm(&tx_adsi, "This is the message");
                    break;
                case DLL_SMS_P2_ENQ:
                    l += 2;
                    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_ACK0, NULL, 0);
                    adsi_put_message(&tx_adsi, tx_msg, tx_len);
                    break;
                case DLL_SMS_P2_REL:
                    l += 2;
                    tx_len = adsi_tl_add_field(&tx_adsi, tx_msg, -1, DLL_SMS_P2_ACK0, NULL, 0);
                    adsi_put_message(&tx_adsi, tx_msg, tx_len);
                    break;
                }
            }
        }
    }
    while (l > 0);
    printf("\n");
}
/*- End of function --------------------------------------------------------*/

static int sms_exec(struct ast_channel *chan, void *data)
{
    int res = 0;
    int count = 0;
    int percentflag = 0;
    char msg[256];
    char *vdata;
    char *localident;
    int i;
    adsi_rx_state_t rx_adsi;
    uint8_t adsi_msg[256];
    int adsi_msg_len;

    struct localuser *u;
    struct ast_frame *inf = NULL;
    struct ast_frame outf;

    int original_read_fmt;
    int original_write_fmt;
    
    uint8_t __buf[sizeof(uint16_t)*BLOCK_SIZE + 2*AST_FRIENDLY_OFFSET];
    uint8_t *buf = __buf + AST_FRIENDLY_OFFSET;
    int len;
    extern int errno;

    if (chan == NULL)
    {
        ast_log(LOG_WARNING, "Channel is NULL. Giving up.\n");
        return -1;
    }
    vdata = data;

    /* The next few lines of code parse out the message and header from the input string */
    if (!vdata)
    {
        /* No data implies no filename or anything is present */
        ast_log(LOG_WARNING, "SMS requires an argument (message)\n");
        return -1;
    }
    
    for (i = 0;  vdata[i]  &&  (vdata[i] != ':')  &&  (vdata[i] != '|');  i++)
    {
        if ((vdata[i] == '%')  &&  (vdata[i + 1] == 'd'))
            percentflag = 1;                      /* the wildcard is used */
        
        if (i == strlen(vdata))
        {
            ast_log(LOG_WARNING, "No extension found\n");
            return -1;
        }
        msg[i] = vdata[i];
    }
    msg[i] = '\0';

printf("Message is '%s'\n", msg);
    if (vdata[i] == '|')
    {
        i++;
    }
    /* Done parsing */

    LOCAL_USER_ADD(u);

    if (chan->_state != AST_STATE_UP)
    {
        /* Shouldn't need this, but checking to see if channel is already answered
         * Theoretically asterisk should already have answered before running the app */
        res = ast_answer(chan);
    }
    
    if (!res)
    {
        original_read_fmt = chan->readformat;
        if (original_read_fmt != AST_FORMAT_SLINEAR)
        {
            res = ast_set_read_format(chan, AST_FORMAT_SLINEAR);
            if (res < 0)
            {
                ast_log(LOG_WARNING, "Unable to set to linear read mode, giving up\n");
                return -1;
            }
        }
        original_write_fmt = chan->writeformat;
        if (original_write_fmt != AST_FORMAT_SLINEAR)
        {
            res = ast_set_write_format(chan, AST_FORMAT_SLINEAR);
            if (res < 0)
            {
                ast_log(LOG_WARNING, "Unable to set to linear write mode, giving up\n");
                res = ast_set_read_format(chan, original_read_fmt);
                if (res)
                    ast_log(LOG_WARNING, "Unable to restore read format on '%s'\n", chan->name);
                return -1;
            }
        }
        adsi_tx_init(&tx_adsi, ADSI_STANDARD_CLIP);
        adsi_rx_init(&rx_adsi, ADSI_STANDARD_CLIP, put_adsi_msg, &rx_adsi);
        adsi_msg_len = adsi_create_message(&tx_adsi, adsi_msg);
        if (msg[0] == '\0')
            adsi_put_message(&tx_adsi, adsi_msg, adsi_msg_len);

        while (ast_waitfor(chan, -1) > -1)
        {
            inf = ast_read(chan);
            if (inf == NULL)
            {
                res = -1;
                break;
            }
            if (inf->frametype == AST_FRAME_VOICE)
            {
                adsi_rx(&rx_adsi, inf->data, inf->samples);
                len = adsi_tx(&tx_adsi, (int16_t *) &buf[AST_FRIENDLY_OFFSET], inf->samples);
                if (len)
                {
                    memset(&outf, 0, sizeof(outf));
                    outf.frametype = AST_FRAME_VOICE;
                    outf.subclass = AST_FORMAT_SLINEAR;
                    outf.datalen = len*sizeof(int16_t);
                    outf.samples = len;
                    outf.data = &buf[AST_FRIENDLY_OFFSET];
                    outf.offset = AST_FRIENDLY_OFFSET;
                    if (ast_write(chan, &outf) < 0)
                    {
                        ast_log(LOG_WARNING, "Unable to write frame to channel; %s\n", strerror(errno));
                        break;
                    }
                }
            }
            ast_frfree(inf);
        }
        if (inf == NULL)
        {
            ast_log(LOG_DEBUG, "Got hangup\n");
            res = -1;
        }
        if (original_read_fmt != AST_FORMAT_SLINEAR)
        {
            res = ast_set_read_format(chan, original_read_fmt);
            if (res)
                ast_log(LOG_WARNING, "Unable to restore read format on '%s'\n", chan->name);
        }
        if (original_write_fmt != AST_FORMAT_SLINEAR)
        {
            res = ast_set_write_format(chan, original_write_fmt);
            if (res)
                ast_log(LOG_WARNING, "Unable to restore write format on '%s'\n", chan->name);
        }
    }
    else
    {
        ast_log(LOG_WARNING, "Could not answer channel '%s'\n", chan->name);
    }
    LOCAL_USER_REMOVE(u);
    return res;
}

int unload_module(void)
{
    STANDARD_HANGUP_LOCALUSERS;
    return ast_unregister_application(app);
}

int load_module(void)
{
    return ast_register_application(app, sms_exec, synopsis, descrip);
}

char *description(void)
{
    return tdesc;
}

int usecount(void)
{
    int res;
    STANDARD_USECOUNT(res);
    return res;
}

char *key()
{
    return ASTERISK_GPL_KEY;
}
